﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Diagnostics;
using DataBaseTest;

namespace AudioChatServer
{
    public class KeepaliveManager
    {
        private Thread _keepAliveThread;
        private Dictionary<IPAddress, int> _keepAliveClients;
        private Stopwatch _keepAliveTimer;
        private UdpServer _server;

        public KeepaliveManager(UdpServer server)
        {
            _keepAliveClients = new Dictionary<IPAddress, int>();
            _server = server;
            _keepAliveTimer = new Stopwatch();
        }

        private void keepAliveCallback(object clientIp)
        {
            _keepAliveTimer = new Stopwatch();

            bool run = true;

            while (run)
            {
                if (_keepAliveTimer.IsRunning)
                {
                    if (_keepAliveTimer.Elapsed.Seconds >= 15)
                    {
                        // If we reach this point, we have not received a keepalive message from the IP
                        // in 10 seconds. We can now consider that IP (and it's associated user) as
                        // disconnected.

                        _server.disconnectUser((IPAddress)clientIp);
                        _keepAliveClients.Remove((IPAddress)clientIp); // Remove the IP from the list of IP's to keep alive
                        _keepAliveTimer.Reset(); // Reset the keep alive timer so we start fresh if they reconnect

                        _server.log("client " + (IPAddress)clientIp + " has timed out");
                        _server.formUpdate();

                        run = false;
                    }
                    else
                    {
                        // If the we have received a keepalive from the IP
                        if (_keepAliveClients.Keys.Contains((IPAddress)clientIp))
                        {
                            if (_keepAliveClients[(IPAddress)clientIp] == Protocol.CS_KEEPALIVE)
                            {
                                Console.WriteLine("Keep-alive received from client " + (IPAddress)clientIp);

                                // Send acknowledgement of keepalive
                                byte[] acknowledgement = new byte[2];
                                acknowledgement[0] = 0x01;
                                acknowledgement[1] = Protocol.SC_KEEPALIVE;

                                _server.sendMessage(acknowledgement, (IPAddress)clientIp);
                                Console.WriteLine("sent keepalive to " + (clientIp as IPAddress).ToString());

                                // Reset the keepalive status of the client and the timer
                                _keepAliveClients[(IPAddress)clientIp] = Protocol.CS_SUCCESS;
                                _keepAliveTimer.Reset();

                                Thread.Sleep(3000);
                            }
                            else if (_keepAliveClients[(IPAddress)clientIp] == Protocol.CS_USERLOGOUT)
                            {
                                //_server.disconnectUser((IPAddress)clientIp);
                                _keepAliveClients.Remove((IPAddress)clientIp);
                                _keepAliveTimer.Reset();
                                Console.WriteLine("   removing user " + (clientIp as IPAddress).ToString());

                                _server.log("client " + (IPAddress)clientIp + " has logged out");
                                _server.formUpdate();

                                run = false;
                            }
                        }
                    }
                }
                else
                {
                    _keepAliveTimer.Restart(); // Start the timer if it is not running
                }
            }
        }

        public bool forgetUser(IPAddress ip)
        {
            try
            {
                if (_keepAliveClients.Keys.Contains(ip))
                {
                    Console.WriteLine("   forgot user " + ip.ToString());
                    _keepAliveClients[ip] = Protocol.CS_USERLOGOUT;
                    //_keepAliveClients.Remove(ip);
                    return true;
                }

                return false;
            }
            catch (Exception e)
            {
                Console.WriteLine("KeepaliveManager.forgetUser() failed: " + e.Message);
                return false;
            }
        }

        public bool keepNewUser(IPAddress ip)
        {
            try
            {
                if (!_keepAliveClients.Keys.Contains(ip))
                {
                    _keepAliveThread = new Thread(keepAliveCallback);
                    _keepAliveThread.Name = "Keep Alive Thread: " + ip.ToString();
                    _keepAliveThread.IsBackground = true;

                    _keepAliveThread.Start(ip);
                    _keepAliveClients.Add(ip, Protocol.CS_KEEPALIVE);

                    _server.log("Keeping new IP: " + ip.ToString());
                    _server.formUpdate();

                    return true;
                }

                return false;
            }
            catch (Exception e)
            {
                Console.WriteLine("KeepaliveManager.keepNewUser() failed: " + e.Message);
                return false;
            }
        }

        public bool sustainUser(IPAddress ip)
        {
            try
            {
                if (_keepAliveClients.Keys.Contains(ip))
                {
                    _keepAliveClients[ip] = Protocol.CS_KEEPALIVE;
                    return true;
                }

                return false;
            }
            catch (Exception e)
            {
                Console.WriteLine("KeepaliveManager.sustainUser() failed: " + e.Message);
                return false;
            }
        }
    }
}
