﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Diagnostics;
using DataBaseTest;

namespace AudioChatServer
{
    public class RoomManager
    {
        UdpServer _server;
        DatabaseInterface _db;
        List<Chatroom> _rooms;

        public RoomManager(UdpServer server, DatabaseInterface dbInterface)
        {
            _server = server;
            _db = dbInterface;
            _rooms = new List<Chatroom>();

            _server.addCallback(Protocol.CS_ADDMEMBER, onAddMember);
            _server.addCallback(Protocol.CS_REMOVEMEMBER, onRemoveMember);
            _server.addCallback(Protocol.CS_ROOMCREATE, onRoomCreate);
            _server.addCallback(Protocol.CS_USERJOINREQUEST, onJoinRequest);
            _server.addCallback(Protocol.CS_USERLEAVEREQUEST, onLeaveRequest);
            _server.addCallback(Protocol.CS_ROOMLISTREQUEST, onRoomListRequest);
        }

        public void initConnection()
        {
            _rooms = _db.GetChatroomList();
        }

        private int onAddMember(IPAddress from, byte[] bytes)
        {
            try
            {
                if (_server.isIpLoggedIn(from))
                {
                    // add member to chatroom through database
                    string roomName = _server.strFromMsg(bytes, 2, Protocol.SIZE_ROOMNAME);
                    Chatroom room = roomWithName(roomName);

                    // Make sure the requesting client is the owner of the room
                    if (room != null && room.owner == _server.userWithIP(from).username)
                    {
                        // Get the usernames from the bytes and put them into a string list
                        List<string> usersToAdd = new List<string>();
                        int usersAdded = 0;

                        for (int i = 3 + (int)Protocol.SIZE_ROOMNAME; i < bytes.Length; i += (int)Protocol.SIZE_USERNAME)
                        {
                            string username = Encoding.ASCII.GetString(bytes, i, (int)Protocol.SIZE_USERNAME);

                            if (username != "")
                            {
                                // Trim any excess null terminators
                                username = username.Remove(username.IndexOf('\0'));

                                Console.WriteLine("     adding " + username);
                                int r = _db.RoomAddUser(roomName, username);
                                switch (r)
                                {
                                    case -1:
                                        _server.log("An error occured in adding user " + username + " to room " + roomName + ".");
                                        break;
                                    case 0:
                                        _server.log("Bad stuff happened with adding member.");
                                        break;
                                    case 1:
                                        if (room.addMember(username))
                                        {
                                            // Give the newly added user their new list of rooms
                                            User addedUser = _server.userWithName(username);

                                            if (_server.isUserLoggedIn(addedUser))
                                            {
                                                byte[] msg = { 0x01, Protocol.CS_ROOMLISTREQUEST };
                                                onRoomListRequest(_server.ipOfUser(addedUser), msg);
                                            }

                                            _server.log("User " + username + " is now a member of room " + roomName);
                                            _server.formUpdate();
                                            usersAdded++;
                                            break;
                                        }
                                        else
                                        {
                                            _server.log("Failed to add user " + username + " to room " + roomName);
                                            break;
                                        }
                                    default:
                                        _server.log("The database returned an unrecognized code " + r +
                                            " when attempting to add user " + username + " to room " + roomName);
                                        break;
                                }
                            }
                        }

                        byte[] returnMsg = { 0x01, Protocol.SC_ADDMEMBER_SUCCESS, (byte)usersAdded };
                        _server.sendMessage(returnMsg, from);

                        return Protocol.SC_SUCCESS;
                    }

                    return Protocol.SC_ADDMEMBER_FAIL;
                }
                else
                {
                    Console.WriteLine(from.ToString() + " is not seen as logged in. (addmember)");
                    return Protocol.SC_ADDMEMBER_FAIL;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("RoomManager.onAddMember() failed: " + e.Message);
                return Protocol.SC_ADDMEMBER_FAIL;
            }
        }

        private int onRemoveMember(IPAddress from, byte[] bytes)
        {
            try
            {
                if (_server.isIpLoggedIn(from))
                {
                    // add member to chatroom through database
                    string roomName = _server.strFromMsg(bytes, 2, Protocol.SIZE_ROOMNAME);
                    Chatroom room = roomWithName(roomName);

                    // Make sure the requesting client is the owner of the room
                    if (room != null && room.owner == _server.userWithIP(from).username)
                    {
                        // Get the usernames from the bytes and put them into a string list
                        List<string> usersToAdd = new List<string>();
                        int usersRemoved = 0;

                        for (int i = 3 + (int)Protocol.SIZE_ROOMNAME; i < bytes.Length; i += (int)Protocol.SIZE_USERNAME)
                        {
                            string username = Encoding.ASCII.GetString(bytes, i, (int)Protocol.SIZE_USERNAME);

                            if (username != "")
                            {
                                // Trim any excess null terminators
                                username = username.Remove(username.IndexOf('\0'));

                                Console.WriteLine("     removing " + username);
                                int r = _db.RoomRemoveUser(roomName, username);
                                switch (r)
                                {
                                    case -1:
                                        _server.log("An error occured in removing user " +
                                            username + " from room " + roomName + ".");
                                        break;
                                    case 0:
                                        _server.log("Bad stuff happened with removing member.");
                                        break;
                                    case 1:
                                        if (room.removeMember(username))
                                        {
                                            // Give the newly added user their new list of rooms
                                            User removedUser = _server.userWithName(username);

                                            if (_server.isUserLoggedIn(removedUser))
                                            {
                                                byte[] msg = { 0x01, Protocol.CS_ROOMLISTREQUEST };
                                                onRoomListRequest(_server.ipOfUser(removedUser), msg);
                                            }

                                            _server.log("User " + username + " is no longer a member of room " + roomName);
                                            _server.formUpdate();
                                            usersRemoved++;
                                            break;
                                        }
                                        else
                                        {
                                            _server.log("Failed to remove user " + username + " from room " + roomName);
                                            break;
                                        }
                                    default:
                                        _server.log("The database returned an unrecognized code " + r +
                                            " when attempting to remove user " + username + " from room " + roomName);
                                        break;
                                }
                            }
                        }

                        byte[] returnMsg = { 0x01, Protocol.SC_REMOVEMEMBER_SUCCESS, (byte)usersRemoved };
                        _server.sendMessage(returnMsg, from);

                        return Protocol.SC_SUCCESS;
                    }

                    return Protocol.SC_REMOVEMEMBER_FAIL;
                }
                else
                {
                    Console.WriteLine(from.ToString() + " is not seen as logged in. (removemember)");
                    return Protocol.SC_REMOVEMEMBER_FAIL;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("RoomManager.onRemoveMember() failed: " + e.Message);
                return Protocol.SC_REMOVEMEMBER_FAIL;
            }
        }

        private int onRoomCreate(IPAddress from, byte[] bytes)
        {
            try
            {
                if (_server.isIpLoggedIn(from))
                {
                    string roomName = _server.strFromMsg(bytes, 2, Protocol.SIZE_ROOMNAME);
                    User user = _server.userWithIP(from);
                    Chatroom roomToAdd = _db.CreateRoom(roomName, _server.userWithIP(from).username);

                    if (user != null && roomToAdd != null)
                    {
                        _rooms.Add(roomToAdd);
                        _server.log(user.username + " created room " + roomName);
                        _server.formUpdate();

                        return Protocol.SC_ROOMCREATE_SUCCESS;
                    }
                    return Protocol.SC_ROOMCREATE_FAIL;
                }
                else
                {
                    return Protocol.SC_ROOMCREATE_FAIL;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("RoomManager.onRoomCreate() failed: " + e.Message);
                return Protocol.SC_ROOMCREATE_FAIL;
            }
        }

        private int onJoinRequest(IPAddress from, byte[] bytes)
        {
            Console.WriteLine("join request potentially unstable");
            try
            {
                if (_server.isIpLoggedIn(from))
                {
                    // add member to chatroom through database
                    string roomName = _server.strFromMsg(bytes, 2, Protocol.SIZE_ROOMNAME);

                    Chatroom room = roomWithName(roomName);
                    User user = _server.userWithIP(from);

                    if (room != null && user != null && user.activeRoom == null)
                    {
                        if (room.userJoin(user.username))
                        {
                            user.activeRoom = room;

                            _server.log(user.username + " has joined room " + room.roomName);
                            _server.formUpdate();

                            List<byte> msg = new List<byte>();
                            msg.Add(0x01);
                            msg.Add(Protocol.SC_USERJOIN);

                            byte[] usr = new byte[Protocol.SIZE_ROOMNAME];
                            ASCIIEncoding.ASCII.GetBytes(user.username).ToList().CopyTo(usr);
                            msg.AddRange(usr.ToList());

                            foreach (string u in room.activeMembers)
                            {
                                _server.sendMessage(msg.ToArray(), _server.userWithName(u).ipAddress);
                            }

                            return Protocol.SC_USERJOINREQUEST_SUCCESS;
                        }
                        else
                        {
                            return Protocol.SC_USERJOINREQUEST_FAIL;
                        }
                    }

                    return Protocol.SC_USERJOINREQUEST_FAIL;
                }
                else
                {
                    return Protocol.SC_USERJOINREQUEST_FAIL;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("RoomManager.onJoinRequest() failed: " + e.Message);
                return Protocol.SC_USERJOINREQUEST_FAIL;
            }
        }

        private int onLeaveRequest(IPAddress from, byte[] bytes)
        {
            Console.WriteLine("leave request untested");
            try
            {
                if (_server.isIpLoggedIn(from))
                {
                    User user = _server.userWithIP(from);

                    if (user != null && user.activeRoom != null)
                    {
                        string room = user.activeRoom.roomName;
                        if (user.activeRoom.userLeave(user.username))
                        {
                            _server.log(user.username + " has left room " + room);
                            _server.formUpdate();

                            List<byte> msg = new List<byte>();
                            msg.Add(0x01);
                            msg.Add(Protocol.SC_USERLEAVE);

                            byte[] usr = new byte[Protocol.SIZE_ROOMNAME];
                            ASCIIEncoding.ASCII.GetBytes(user.username).ToList().CopyTo(usr);
                            msg.AddRange(usr.ToList());

                            foreach (string u in user.activeRoom.activeMembers)
                            {
                                _server.sendMessage(msg.ToArray(), _server.userWithName(u).ipAddress);
                            }

                            user.activeRoom = null;
                            return Protocol.SC_USERLEAVEREQUEST_SUCCESS;
                        }
                        else
                        {
                            return Protocol.SC_USERLEAVEREQUEST_FAIL;
                        }
                    }

                    return Protocol.SC_USERLEAVEREQUEST_FAIL;
                }
                else
                {
                    return Protocol.SC_USERLEAVEREQUEST_FAIL;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("RoomManager.onLeaveRequest() failed: " + e.Message);
                return Protocol.SC_USERLEAVEREQUEST_FAIL;
            }
        }

        private int onRoomListRequest(IPAddress ip, byte[] bytes)
        {
            try
            {
                User user = _server.userWithIP(ip);

                if (user != null)
                {
                    List<string> membership = roomMembership(user.username);

                    List<byte> msg = new List<byte>();
                    msg.Add(0x01);
                    msg.Add(Protocol.SC_ROOMLIST);
                    msg.Add(Convert.ToByte(membership.Count));

                    foreach (string r in membership)
                    {
                        byte[] room = new byte[Protocol.SIZE_ROOMNAME];
                        ASCIIEncoding.ASCII.GetBytes(r).ToList().CopyTo(room);
                        msg.AddRange(room.ToList());
                    }

                    _server.sendMessage(msg.ToArray(), ip);

                    return Protocol.SC_SUCCESS;
                }

                return Protocol.SC_FAILURE;
            }
            catch (Exception e)
            {
                Console.WriteLine("RoomManager.onRoomListRequest() failed: " + e.Message);
                return Protocol.SC_FAILURE;
            }
        }



        public Chatroom roomWithName(string roomName)
        {
            foreach (Chatroom r in _rooms)
            {
                if (r.roomName == roomName)
                {
                    return r;
                }
            }

            return null;
        }

        public List<string> roomMembership(string username)
        {
            List<string> result = new List<string>();

            foreach (Chatroom room in _db.GetUserChatroomAccessList(username))
            {
                result.Add(room.roomName);
            }

            return result;
        }

        public bool isUserInRoom(User user, Chatroom room)
        {
            if (user != null && room != null)
            {
                return room.isUserAMember(user.username);
            }

            return false;
        }

        public bool isUserInRoom(string userName, string roomName)
        {
            return isUserInRoom(_server.userWithName(userName), roomWithName(roomName));
        }
    }
}
