﻿/*
 * Chris Logsdon
 * 
 * Credits:
 * http://stackoverflow.com/questions/10538862/c-sharp-udp-server-with-multiple-users
 */

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Diagnostics;
using DataBaseTest;

namespace AudioChatServer
{
    public class State
    {
        public UdpClient listener;
        public IPEndPoint ipEndPoint;
    }

    public class UdpServer
    {
        private IPAddress _ip;
        private UdpClient _listener, _sender;
        private State _stateSend, _stateRecv;
        private bool _isListening;

        public delegate void onClientMessageReceive(string msg);
        private onClientMessageReceive _onReceiveCallback;

        public delegate int messageReceiveCallback(IPAddress from, byte[] bytes);
        Dictionary<int, messageReceiveCallback> messageCallbacks;

        private DatabaseInterface _dbInterface;

        public Queue<string> logMessages;
        public MainForm.UpdateDelegate formUpdate;

        KeepaliveManager keepaliveManager;
        UserManager userManager;
        RoomManager roomManager;
        InteractionManager interactionManager;

        public UdpServer(onClientMessageReceive onReceiveCallback, MainForm.UpdateDelegate formUpdate)
        {
            this._onReceiveCallback = onReceiveCallback;

            _ip = findIPAddress();

            _isListening = false;
            
            logMessages = new Queue<string>();
            
            this.formUpdate = formUpdate;


            _dbInterface = new DatabaseInterface("gladev.aukware.com", "test", "", "chatty");

            // Keepalive
            keepaliveManager = new KeepaliveManager(this);

            // Server actions
            messageCallbacks = new Dictionary<int, messageReceiveCallback>();

            addCallback(Protocol.CS_KEEPALIVE, onKeepalive);
            addCallback(Protocol.CS_USERCONNECT, onUserConnect);
            addCallback(Protocol.CS_FAILURE, onFailure);
            addCallback(Protocol.CS_SUCCESS, onSuccess);

            userManager = new UserManager(this, _dbInterface);
            roomManager = new RoomManager(this, _dbInterface);
            interactionManager = new InteractionManager(this);
        }

        public bool startListening(int portNumber)
        {
            try
            {
                int port = portNumber;

                _listener = new UdpClient(3210);
                _stateRecv = new State();
                _stateRecv.listener = _listener;
                _stateRecv.ipEndPoint = new IPEndPoint(IPAddress.Any, 3210);

                _sender = new UdpClient();
                _stateSend = new State();
                _stateSend.listener = _sender;
                _stateSend.ipEndPoint = new IPEndPoint(_ip, port);

                if (!_dbInterface.OpenConnection())
                {
                    logMessages.Enqueue(dateNowStr() + " Connection to the database failed.");
                }

                roomManager.initConnection();

                _listener.BeginReceive(new AsyncCallback(receiveCallback), _stateRecv);

                _isListening = true;

                formUpdate();

                return true;
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                _isListening = false;
                return false;
            }
        }

        public void stopListening()
        {
            try
            {
                if (_listener != null)
                {
                    _listener.Client.Close();
                    //_listener.Close();
                    _dbInterface.CloseConnection();
                    userManager.ConnectedUsers.Clear();
                    _isListening = false;

                    formUpdate();
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("close connection failed: " + e.Message);
            }
        }

        private void receiveCallback(IAsyncResult ar)
        {
            try
            {
                State s = (State)(ar.AsyncState);

                byte[] received = _listener.EndReceive(ar, ref _stateRecv.ipEndPoint);
                string text = Encoding.ASCII.GetString(received);

                if (received[0] == 0x01)
                {
                    if (messageCallbacks.Keys.Contains(received[1]))
                    {
                        byte[] reply = new byte[2];
                        reply[0] = 0x01;
                        reply[1] = Convert.ToByte(messageCallbacks[received[1]](_stateRecv.ipEndPoint.Address, received));

                        sendMessage(reply, _stateRecv.ipEndPoint.Address);
                    }
                    else
                    {
                        Console.WriteLine("No callback assigned to receive code " + received[1]);
                    }
                }

                _onReceiveCallback(text);

                s.listener.BeginReceive(new AsyncCallback(receiveCallback), _stateRecv);
            }
            catch (Exception e)
            {
                Console.WriteLine("receive callback failed: " + e.Message);
            }
        }

        public bool disconnectUser(IPAddress ip)
        {
            if (userManager.ConnectedUsers.Keys.Contains(ip))
            {
                User user = userWithIP(ip);
                if (user != null)
                {
                    if (user.activeRoom != null)
                    {
                        foreach (string u in user.activeRoom.activeMembers)
                        {
                            List<byte> msg = new List<byte>();
                            msg.Add(0x01);
                            msg.Add(Protocol.SC_USERLEAVE);

                            byte[] usr = new byte[Protocol.SIZE_USERNAME];
                            ASCIIEncoding.ASCII.GetBytes(user.username).ToList().CopyTo(usr);
                            msg.AddRange(usr.ToList());

                            sendMessage(msg.ToArray(), userWithName(u).ipAddress);
                        }

                        user.activeRoom.userLeave(user.username);
                    }

                    foreach (User u in ConnectedUsers.Values)
                    {
                        if (ipOfUser(u) != ip)
                        {
                            byte[] msg = { 0x01, Protocol.SC_USERLIST, 0x01 };
                            sendMessage(msg, ipOfUser(u));
                        }
                    }
                }
                if (!userManager.ConnectedUsers.Remove(ip))
                {
                    return false;
                }

                return keepaliveManager.forgetUser(ip);
            }

            return false;
        }

        public void log(string msg)
        {
            logMessages.Enqueue(dateNowStr() + "\n \t" + msg + "\n");
        }

        public bool sendMessage(byte[] message, IPAddress ipTo, int msgSize = -1)
        {
            try
            {
                if (msgSize <= 0)
                {
                    _sender.Send(message, message.Length, new IPEndPoint(ipTo, portNumberSend));
                }
                else
                {
                    _sender.Send(message, msgSize, new IPEndPoint(ipTo, portNumberSend));
                }
                return true;
            }
            catch (Exception e)
            {
                Console.WriteLine("UdpServer.sendMessage() failed: " + e.Message);
                return false;
            }
        }

        public void addCallback(int messageCode, messageReceiveCallback callback)
        {
            try
            {
                messageCallbacks.Add(messageCode, callback);
            }
            catch (Exception e)
            {
                Console.WriteLine("failed to add callback: " + e.Message);
            }
        }

        public void removeCallback(int messageCode, messageReceiveCallback callback)
        {
            try
            {
                if (messageCallbacks.ContainsKey(messageCode) && messageCallbacks.ContainsValue(callback))
                {
                    messageCallbacks.Remove(messageCode);
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("failed to remove callback: " + e.Message);
            }
        }


#region Utility Functions

        public string dateNowStr()
        {
            return "(" + DateTime.Now.Hour.ToString() + ":" +
                    DateTime.Now.Minute.ToString() + ":" +
                    DateTime.Now.Second.ToString() + ":" +
                    DateTime.Now.Millisecond.ToString() + ")";
        }

        public string strFromMsg(byte[] bytes, uint startIndex, uint stringByteLength)
        {
            string result = "";

            for (uint i = 0; i < stringByteLength; i++)
            {
                char c = Convert.ToChar(bytes[startIndex + i]);

                if (c == '\0')
                {
                    break;
                }
                result += c;
            }

            return result;
        }

        public IPAddress findIPAddress()
        {
            IPAddress result = IPAddress.Parse("127.0.0.1");

            // Grab the computer's IP address if possible. Otherwise, use local host
            try
            {
                IPHostEntry hostEntry = Dns.GetHostByName(Dns.GetHostName());
                foreach (IPAddress ip in hostEntry.AddressList)
                {
                    result = ip;
                    break;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
            }

            return result;
        }

        public IPAddress setIPAddress(string newIP)
        {
            try
            {
                _ip = IPAddress.Parse(newIP);
                return _ip;
            }
            catch (Exception e)
            {
                MessageBox.Show(newIP + " is an invalid IP Address.");
                Console.WriteLine(e.Message);
                return null;
            }
        }


        public User userFromDatabase(string username)
        {
            return userManager.userFromDatabase(username);
        }

        public IPAddress ipOfUser(User user)
        {
            return userManager.ipOfUser(user);
        }

        public IPAddress ipOfUser(string username)
        {
            return userManager.ipOfUser(username);
        }

        public bool isUserLoggedIn(User user)
        {
            return userManager.isUserLoggedIn(user);
        }

        public bool isIpConnected(IPAddress ip)
        {
            return userManager.isIpConnected(ip);
        }

        public bool isIpLoggedIn(IPAddress ip)
        {
            return userManager.isIpLoggedIn(ip);
        }

        public User userWithIP(IPAddress ip)
        {
            return userManager.userWithIP(ip);
        }

        public User userWithName(string username)
        {
            return userManager.userWithName(username);
        }

        public Chatroom roomOfUser(IPAddress ip)
        {
            if (userWithIP(ip) != null)
            {
                return userWithIP(ip).activeRoom;
            }

            return null;
        }

        public Chatroom roomOfUser(string username)
        {
            if (userWithName(username) != null)
            {
                return userWithName(username).activeRoom;
            }

            return null;
        }

        public Chatroom roomOfUser(User user)
        {
            if (user != null)
            {
                return user.activeRoom;
            }

            return null;
        }

        public List<string> roomMembership(string username)
        {
            return roomManager.roomMembership(username);
        }

#endregion


#region Properties

        public string ipAddress
        {
            get { return _ip.ToString(); }
        }

        public int portNumberListen
        {
            get { return _stateRecv.ipEndPoint.Port; }
        }

        public int portNumberSend
        {
            get { return _stateSend.ipEndPoint.Port; }
        }

        public bool isListening
        {
            get { return _isListening; }
        }

        public Dictionary<IPAddress, User> ConnectedUsers
        {
            get { return userManager.ConnectedUsers; }
        }

        public List<User> UsersInDatabase
        {
            get { return _dbInterface.GetUserList(); }
        }

        public List<Chatroom> RoomsInDatabase
        {
            get
            {
                //MessageBox.Show(_dbInterface.GetChatroomList().Count.ToString());
                return _dbInterface.GetChatroomList();
            }
        }

#endregion


#region Server Actions to Client Messages

        private int onFailure(IPAddress ip, byte[] bytes)
        {
            MessageBox.Show("Not implemented");
            try
            {
                return Protocol.SC_SUCCESS;
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                return Protocol.SC_FAILURE;
            }
        }

        private int onSuccess(IPAddress ip, byte[] bytes)
        {
            MessageBox.Show("Not implemented");
            try
            {
                return Protocol.SC_SUCCESS;
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                return Protocol.SC_FAILURE;
            }
        }

        private int onKeepalive(IPAddress ip, byte[] bytes)
        {
            try
            {
                if (isIpConnected(ip))
                {
                    if (keepaliveManager.sustainUser(ip))
                    {
                        return Protocol.SC_SUCCESS;
                    }

                    return Protocol.SC_FAILURE;
                }
                else
                {
                    return Protocol.SC_FAILURE;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                return Protocol.SC_FAILURE;
            }
        }

        private int onUserConnect(IPAddress ip, byte[] bytes)
        {
            try
            {
                if (!isIpConnected(ip))
                {
                    keepaliveManager.keepNewUser(ip);

                    userManager.ConnectedUsers.Add(ip, null);

                    log(ip.ToString() + " has connected.");

                    formUpdate();
                    return Protocol.SC_USERCONNECT_SUCCESS;
                }
                else
                {
                    if (disconnectUser(ip))
                    {
                        log(ip.ToString() + " has disconnected");

                        keepaliveManager.keepNewUser(ip);

                        userManager.ConnectedUsers.Add(ip, null);

                        log(ip.ToString() + " has connected.");

                        formUpdate();
                        return Protocol.SC_USERCONNECT_SUCCESS;
                    }

                    return Protocol.SC_USERCONNECT_FAIL;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("IP connect failed: " + e.Message);

                return Protocol.SC_USERCONNECT_FAIL;
            }
        }
#endregion

    }
}