﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Diagnostics;
using DataBaseTest;

namespace AudioChatServer
{
    public class UserManager
    {
        UdpServer _server;
        DatabaseInterface _db;
        private Dictionary<IPAddress, User> _connectedUsers;

        public UserManager(UdpServer server, DatabaseInterface db)
        {
            _server = server;
            _db = db;

            _connectedUsers = new Dictionary<IPAddress, User>();

            _server.addCallback(Protocol.CS_USERLOGIN, onUserLogin);
            _server.addCallback(Protocol.CS_USERLOGOUT, onUserLogout);
            _server.addCallback(Protocol.CS_USERREGISTER, onUserRegister);
            _server.addCallback(Protocol.CS_USERLISTREQUEST, onUserListRequest);
        }

        private int onUserLogin(IPAddress ip, byte[] bytes)
        {
            try
            {
                if (isIpConnected(ip) && !isIpLoggedIn(ip))
                {
                    string username = _server.strFromMsg(bytes, 2, Protocol.SIZE_USERNAME);
                    string password = _server.strFromMsg(bytes, 2 + Protocol.SIZE_USERNAME, Protocol.SIZE_PASSWORD_HASH);

                    switch (_db.CheckUserExists(username))
                    {
                        case -1:
                            _server.log("AN ERROR OCCURRED IN CHECKING EXISTENCE OF USER " + username);
                            return Protocol.SC_USERLOGIN_FAIL;
                        case 0:
                            Console.WriteLine("User " + username + " does not exist in the database.");
                            return Protocol.SC_USERLOGIN_FAIL;
                        case 1:
                            break;
                        default:
                            return Protocol.SC_USERLOGIN_FAIL;
                    }

                    switch (_db.CheckPassword(username, password))
                    {
                        case -1:
                            _server.log("AN ERROR OCCURRED IN CHECKING PASSWORD OF USER " + username);
                            return Protocol.SC_USERLOGIN_FAIL;
                        case 0:
                            Console.WriteLine("The password is incorrect.");
                            return Protocol.SC_USERLOGIN_FAIL;
                        case 1:
                            _connectedUsers[ip] = _server.userFromDatabase(username);
                            _connectedUsers[ip].ipAddress = ip;

                            _server.log(_connectedUsers[ip].username + " has LOGGED IN as " + ip.ToString());

                            foreach (User u in _connectedUsers.Values)
                            {
                                if (ipOfUser(u) != ip)
                                {
                                    byte[] msg = { 0x01, Protocol.SC_USERLIST, 0x01 };
                                    _server.sendMessage(msg, ipOfUser(u));
                                }
                            }

                            _server.formUpdate();

                            return Protocol.SC_USERLOGIN_SUCCESS;
                        default:
                            return Protocol.SC_USERLOGIN_FAIL;
                    }
                }

                Console.WriteLine(ip.ToString() + " is not seen as connected by the server.");
                return Protocol.SC_USERLOGIN_FAIL;
            }
            catch (Exception e)
            {
                Console.WriteLine("User LOGIN failed: " + e.Message);
                return Protocol.SC_USERLOGIN_FAIL;
            }
        }

        private int onUserLogout(IPAddress ip, byte[] bytes)
        {
            try
            {
                if (_server.disconnectUser(ip))
                {
                    _server.log(ip.ToString() + " has disconnected");
                    _server.formUpdate();

                    return Protocol.SC_USERLOGOUT_SUCCESS;
                }

                return Protocol.SC_USERLOGOUT_FAIL;
            }
            catch (Exception e)
            {
                Console.WriteLine("User LOGOUT failed: " + e.Message);
                return Protocol.SC_USERLOGOUT_FAIL;
            }
        }

        private int onUserRegister(IPAddress ip, byte[] bytes)
        {
            try
            {
                if (isIpConnected(ip))
                {
                    string username = _server.strFromMsg(bytes, 2, Protocol.SIZE_USERNAME);
                    string password = _server.strFromMsg(bytes, 2 + Protocol.SIZE_USERNAME, Protocol.SIZE_PASSWORD_HASH);

                    switch (_db.CheckUserExists(username))
                    {
                        case -1:
                            _server.log("AN ERROR OCCURRED IN CREATING USER " + username);
                            return Protocol.SC_USERREGISTER_FAIL;
                        case 0:
                            if (_db.CreateUser(username, password))
                            {
                                _server.log("User " + username + " successfully created.");

                                _server.formUpdate();
                                return Protocol.SC_USERREGISTER_SUCCESS;
                            }
                            else
                            {
                                _server.log("USER " + username + " CREATION FAILED.");
                                return Protocol.SC_USERREGISTER_FAIL;
                            }
                        case 1:
                            Console.WriteLine("User " + username + " already exists in the database.");
                            return Protocol.SC_USERREGISTER_FAIL;
                        default:
                            Console.WriteLine("Unrecognized code returned from DB when trying to register user.");
                            return Protocol.SC_USERREGISTER_FAIL;
                    }
                }
                else
                {
                    Console.WriteLine("Could not register user because IP is not logged in.");
                    return Protocol.SC_USERREGISTER_FAIL;
                }
            }
            catch (Exception e)
            {
                Console.WriteLine("User registration failed: " + e.Message);
                return Protocol.SC_FAILURE;
            }
        }

        private int onUserListRequest(IPAddress ip, byte[] bytes)
        {
            try
            {
                User user = _server.userWithIP(ip);

                if (user != null)
                {
                    List<string> users = new List<string>();

                    if (bytes[2] == 0x01)
                    {
                        foreach (User u in _connectedUsers.Values)
                        {
                            users.Add(u.username);
                        }
                    }
                    else if (bytes[2] == 0x02)
                    {
                        if (user.activeRoom != null)
                        {
                            foreach (string u in user.activeRoom.activeMembers)
                            {
                                users.Add(u);
                            }
                        }
                    }
                    else if (bytes[2] == 0x03)
                    {
                        if (user.activeRoom != null && user.activeRoom.owner == user.username)
                        {
                            foreach (User u in _server.UsersInDatabase)
                            {
                                if (!user.activeRoom.members.Contains(u.username))
                                {
                                    users.Add(u.username);
                                }
                            }
                        }
                    }
                    else if (bytes[2] == 0x04)
                    {
                        if (user.activeRoom != null && user.activeRoom.owner == user.username)
                        {
                            foreach (User u in _server.UsersInDatabase)
                            {
                                if (u.username != user.username)
                                {
                                    if (user.activeRoom.members.Contains(u.username))
                                    {
                                        users.Add(u.username);
                                    }
                                }
                            }
                        }
                    }
                    else
                    {
                        foreach (User u in _server.UsersInDatabase)
                        {
                            users.Add(u.username);
                        }
                    }

                    List<byte> msg = new List<byte>();
                    msg.Add(0x01);
                    msg.Add(Protocol.SC_USERLIST);
                    msg.Add(bytes[2]);
                    msg.Add(Convert.ToByte(users.Count));

                    foreach (string u in users)
                    {
                        byte[] usr = new byte[Protocol.SIZE_USERNAME];
                        ASCIIEncoding.ASCII.GetBytes(u).ToList().CopyTo(usr);
                        msg.AddRange(usr.ToList());
                    }

                    _server.sendMessage(msg.ToArray(), ip);

                    return Protocol.SC_SUCCESS;
                }

                return Protocol.SC_FAILURE;
            }
            catch (Exception e)
            {
                Console.WriteLine("UserManager.onUserListRequest() failed: " + e.Message);
                return Protocol.SC_FAILURE;
            }
        }

        public bool isIpConnected(IPAddress ip)
        {
            return _connectedUsers.Keys.Contains(ip);
        }

        public bool isIpLoggedIn(IPAddress ip)
        {
            return isIpConnected(ip) && _connectedUsers[ip] != null;
        }

        public User userWithIP(IPAddress ip)
        {
            if (_connectedUsers.ContainsKey(ip))
            {
                return _connectedUsers[ip];
            }

            return null;
        }

        public IPAddress ipOfUser(User user)
        {
            foreach (User u in _connectedUsers.Values)
            {
                if (u != null && user != null)
                {
                    if (u.username == user.username)
                    {
                        return u.ipAddress;
                    }
                }
            }

            return null;
        }

        public IPAddress ipOfUser(string username)
        {
            return ipOfUser(userWithName(username));
        }

        public bool isUserLoggedIn(User user)
        {
            foreach (User u in _connectedUsers.Values)
            {
                if (u != null && user != null)
                {
                    if (u.username == user.username)
                    {
                        return true;
                    }
                }
            }

            return false;
        }

        public User userWithName(string username)
        {
            foreach (User u in _connectedUsers.Values)
            {
                if (u != null)
                {
                    if (u.username == username)
                    {
                        return u;
                    }
                }
            }

            return null;
        }

        public User userFromDatabase(string username)
        {
            foreach (User u in _db.GetUserList())
            {
                if (u.username == username)
                {
                    return u;
                }
            }

            return null;
        }

        public Dictionary<IPAddress, User> ConnectedUsers
        {
            get { return _connectedUsers; }
        }
    }
}
